﻿using EonaCat.HID.EventArguments;
using EonaCat.HID.Managers;
using EonaCat.HID.Managers.Linux;
using System;
using System.Collections.Generic;
using System.IO;
using System.Runtime.InteropServices;
using System.Threading;
using System.Threading.Tasks;

namespace EonaCat.HID
{
    // This file is part of the EonaCat project(s) which is released under the Apache License.
    // See the LICENSE file or go to https://EonaCat.com/license for full license details.

    internal sealed class HidLinux : IHid
    {
        private readonly string _devicePath;
        private FileStream _stream;
        private readonly object _lock = new object();
        private CancellationTokenSource _cts;

        public string DevicePath => _devicePath;
        public ushort VendorId { get; private set; }
        public ushort ProductId { get; private set; }
        public string SerialNumber { get; private set; }
        public string Manufacturer { get; private set; }
        public string ProductName { get; private set; }
        public int InputReportByteLength { get; private set; }
        public int OutputReportByteLength { get; private set; }
        public int FeatureReportByteLength { get; private set; }
        public IDictionary<string, object> Capabilities { get; private set; } = new Dictionary<string, object>();

        public event EventHandler<HidDataReceivedEventArgs> OnDataReceived;
        public event EventHandler<HidErrorEventArgs> OnError;

        public HidLinux(string devicePath)
        {
            _devicePath = devicePath ?? throw new ArgumentNullException(nameof(devicePath));
        }

        internal void Setup()
        {
            Open();
            LoadDeviceInfo();
            LoadReportLengths();
            Close();
        }

        public void Open()
        {
            lock (_lock)
            {
                if (_stream != null)
                {
                    return;
                }

                // Open device for read/write
                var fd = NativeMethods.open(_devicePath, NativeMethods.O_RDWR | NativeMethods.O_NONBLOCK);
                if (fd < 0)
                {
                    OnError?.Invoke(this, new HidErrorEventArgs(this, new IOException($"Failed to open device: {_devicePath}")));
                    return;
                }

                var safeHandle = new Microsoft.Win32.SafeHandles.SafeFileHandle(new IntPtr(fd), ownsHandle: true);
                _stream = new FileStream(safeHandle, FileAccess.ReadWrite, bufferSize: 64, isAsync: false);
            }
        }

        public void Close()
        {
            lock (_lock)
            {
                StopListening();

                if (_stream != null)
                {
                    _stream.Dispose();
                    _stream = null;
                }
            }
        }

        public void Dispose()
        {
            Close();
        }

        private void LoadDeviceInfo()
        {
            string sysPath = GetSysfsDevicePath(_devicePath);
            if (string.IsNullOrEmpty(sysPath))
            {
                return;
            }

            VendorId = ReadHexFileAsUShort(Path.Combine(sysPath, "idVendor"));
            ProductId = ReadHexFileAsUShort(Path.Combine(sysPath, "idProduct"));
            SerialNumber = ReadTextFile(Path.Combine(sysPath, "serial"));
            Manufacturer = ReadTextFile(Path.Combine(sysPath, "manufacturer"));
            ProductName = ReadTextFile(Path.Combine(sysPath, "product"));
        }

        private void LoadReportLengths()
        {
            // As Linux/hidraw interface does not provide direct API for report lengths,
            // We fallback to standard max USB HID report size or fixed size
            // TODO: Maybe we should use hidapi or libusb bindings?

            InputReportByteLength = 64;    // usually max 64 on USB HID
            OutputReportByteLength = 64;
            FeatureReportByteLength = 64;
        }

        private string GetSysfsDevicePath(string devPath)
        {
            // Mapping /dev/hidrawX to sysfs device path: /sys/class/hidraw/hidrawX/device/ 
            var devName = Path.GetFileName(devPath);
            var sysDevicePath = Path.Combine("/sys/class/hidraw", devName, "device");
            if (Directory.Exists(sysDevicePath))
            {
                return sysDevicePath;
            }

            return null;
        }

        private ushort ReadHexFileAsUShort(string path)
        {
            try
            {
                if (File.Exists(path))
                {
                    string text = File.ReadAllText(path).Trim().ToLowerInvariant();
                    if (text.StartsWith("0x"))
                    {
                        text = text.Substring(2);
                    }

                    if (ushort.TryParse(text, System.Globalization.NumberStyles.HexNumber, null, out ushort val))
                    {
                        return val;
                    }
                }
            }
            catch { }
            return 0;
        }

        private string ReadTextFile(string path)
        {
            try
            {
                if (File.Exists(path))
                {
                    return File.ReadAllText(path).Trim();
                }
            }
            catch { }
            return null;
        }

        public async Task WriteOutputReportAsync(byte[] data)
        {
            if (data == null)
            {
                OnError?.Invoke(this, new HidErrorEventArgs(this, new ArgumentNullException(nameof(data))));
                return;
            }

            if (_stream == null)
            {
                OnError?.Invoke(this, new HidErrorEventArgs(this, new InvalidOperationException("Device not open")));
                return;
            }

            await Task.Run(() =>
            {
                lock (_lock)
                {
                    try
                    {
                        _stream.Write(data, 0, data.Length);
                        _stream.Flush();
                    }
                    catch (Exception ex)
                    {
                        OnError?.Invoke(this, new HidErrorEventArgs(this, ex));
                    }
                }
            });
        }

        public async Task<byte[]> ReadInputReportAsync()
        {
            if (_stream == null)
            {
                OnError?.Invoke(this, new HidErrorEventArgs(this, new InvalidOperationException("Device not open")));
                return Array.Empty<byte>();
            }

            return await Task.Run(() =>
            {
                byte[] buffer = new byte[InputReportByteLength];
                int bytesRead = 0;

                lock (_lock)
                {
                    try
                    {
                        bytesRead = _stream.Read(buffer, 0, buffer.Length);
                    }
                    catch (Exception ex)
                    {
                        OnError?.Invoke(this, new HidErrorEventArgs(this, ex));
                    }
                }

                if (bytesRead <= 0)
                {
                    return Array.Empty<byte>();
                }

                byte[] result = new byte[bytesRead];
                Array.Copy(buffer, result, bytesRead);
                return result;
            });
        }

        public async Task SendFeatureReportAsync(byte[] data)
        {
            if (data == null || data.Length == 0)
                throw new ArgumentException("Feature report data must not be null or empty.");

            int fd = NativeMethods.open(_devicePath, NativeMethods.O_RDWR);
            if (fd < 0)
            {
                OnError?.Invoke(this, new HidErrorEventArgs(this, new IOException($"Failed to open device: {_devicePath}")));
                return;
            }

            try
            {
                int size = data.Length;
                IntPtr buffer = Marshal.AllocHGlobal(size);
                Marshal.Copy(data, 0, buffer, size);

                int request = NativeMethods._IOC(NativeMethods._IOC_WRITE, 'H', 0x06, size); // HIDIOCSFEATURE
                int result = NativeMethods.ioctl(fd, request, buffer);
                if (result < 0)
                {
                    OnError?.Invoke(this, new HidErrorEventArgs(this, new IOException("ioctl HIDIOCSFEATURE failed")));
                    return;
                }

                Marshal.FreeHGlobal(buffer);
            }
            finally
            {
                NativeMethods.close(fd);
            }

            await Task.CompletedTask;
        }

        public async Task<byte[]> GetFeatureReportAsync(byte reportId)
        {
            const int maxFeatureSize = 256;
            byte[] buffer = new byte[maxFeatureSize];
            buffer[0] = reportId;

            int fd = NativeMethods.open(_devicePath, NativeMethods.O_RDWR);
            if (fd < 0)
            {
                OnError?.Invoke(this, new HidErrorEventArgs(this, new IOException($"Failed to open device: {_devicePath}")));
                return Array.Empty<byte>();
            }

            try
            {
                IntPtr bufPtr = Marshal.AllocHGlobal(buffer.Length);
                Marshal.Copy(buffer, 0, bufPtr, buffer.Length);

                int request = NativeMethods._IOC(NativeMethods._IOC_READ, 'H', 0x07, buffer.Length); // HIDIOCGFEATURE
                int result = NativeMethods.ioctl(fd, request, bufPtr);
                if (result < 0)
                {
                    OnError?.Invoke(this, new HidErrorEventArgs(this, new IOException("ioctl HIDIOCGFEATURE failed")));
                    return Array.Empty<byte>();
                }

                Marshal.Copy(bufPtr, buffer, 0, buffer.Length);
                Marshal.FreeHGlobal(bufPtr);

                return await Task.FromResult(buffer);
            }
            finally
            {
                NativeMethods.close(fd);
            }
        }

        public async Task StartListeningAsync(CancellationToken cancellationToken)
        {
            if (_stream == null)
            {
                OnError?.Invoke(this, new HidErrorEventArgs(this, new NotSupportedException("Device is not open.")));
                return;
            }

            lock (_lock)
            {
                if (_cts != null)
                {
                    OnError?.Invoke(this, new HidErrorEventArgs(this, new NotSupportedException("Already listening on this device.")));
                    return;
                }

                _cts = CancellationTokenSource.CreateLinkedTokenSource(cancellationToken);
            }

            try
            {
                var token = _cts.Token;

                while (!token.IsCancellationRequested)
                {
                    byte[] buffer = new byte[InputReportByteLength];
                    int bytesRead;

                    try
                    {
                        bytesRead = await _stream.ReadAsync(buffer, 0, buffer.Length, token);
                    }
                    catch (OperationCanceledException)
                    {
                        break;
                    }
                    catch (NotSupportedException)
                    {
                        OnError?.Invoke(this, new HidErrorEventArgs(this, new NotSupportedException("Reading input reports is not supported on this device.")));
                        break; // Exit if reading is not supported
                    }
                    catch (Exception ex)
                    {
                        OnError?.Invoke(this, new HidErrorEventArgs(this, ex));
                        break;
                    }

                    if (bytesRead > 0)
                    {
                        var data = new byte[bytesRead];
                        Array.Copy(buffer, data, bytesRead);

                        OnDataReceived?.Invoke(this, new HidDataReceivedEventArgs(this, data));
                    }
                    else
                    {
                        try
                        {
                            await Task.Delay(5, token);
                        }
                        catch (OperationCanceledException)
                        {
                            break;
                        }
                    }
                }
            }
            finally
            {
                lock (_lock)
                {
                    _cts?.Dispose();
                    _cts = null;
                }
            }
        }

        private void StopListening()
        {
            lock (_lock)
            {
                if (_cts != null)
                {
                    _cts.Cancel();
                    _cts.Dispose();
                    _cts = null;
                }
            }
        }
    }
}