﻿using EonaCat.HID.EventArguments;
using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Threading;
using System.Threading.Tasks;

namespace EonaCat.HID
{
    // This file is part of the EonaCat project(s) which is released under the Apache License.
    // See the LICENSE file or go to https://EonaCat.com/license for full license details.

    internal sealed class HidMac : IHid
    {
        private IntPtr _deviceHandle;

        public HidMac(IntPtr deviceHandle)
        {
            _deviceHandle = deviceHandle;
            CFRetain(_deviceHandle);
        }

        internal void Setup()
        {
            InitializeDeviceInfo();
        }

        public string DevicePath => null;
        public ushort VendorId { get; private set; }
        public ushort ProductId { get; private set; }
        public string SerialNumber { get; private set; }
        public string Manufacturer { get; private set; }
        public string ProductName { get; private set; }

        public int InputReportByteLength { get; private set; } = 64;
        public int OutputReportByteLength { get; private set; } = 64;
        public int FeatureReportByteLength { get; private set; } = 64;

        public IDictionary<string, object> Capabilities { get; private set; } = new Dictionary<string, object>();

        public event EventHandler<HidDataReceivedEventArgs> OnDataReceived;
        public event EventHandler<HidErrorEventArgs> OnError;

        public void Open()
        {
            IOHIDDeviceOpen(_deviceHandle, 0);
        }

        public void Close()
        {
            IOHIDDeviceClose(_deviceHandle, 0);
        }

        public void Dispose()
        {
            StopListening();
            Close();
            CFRelease(_deviceHandle);
            _deviceHandle = IntPtr.Zero;
        }

        public void StopListening()
        {
            _listeningCts?.Cancel();
        }

        public async Task WriteOutputReportAsync(byte[] data)
        {
            if (data is null)
            {
                throw new ArgumentNullException(nameof(data));
            }

            await Task.Run(() =>
            {
                byte reportId = data.Length > 0 ? data[0] : (byte)0;
                IntPtr buffer = Marshal.AllocHGlobal(data.Length);

                try
                {
                    Marshal.Copy(data, 0, buffer, data.Length);
                    int res = IOHIDDeviceSetReport(_deviceHandle, 1, reportId, buffer, data.Length);
                    if (res != 0)
                    {
                        OnError?.Invoke(this, new HidErrorEventArgs(this, new Exception($"IOHIDDeviceSetReport (Output) failed: {res}")));
                        return;
                    }
                }
                finally
                {
                    Marshal.FreeHGlobal(buffer);
                }
            });
        }

        public async Task SendFeatureReportAsync(byte[] data)
        {
            if (data is null)
            {
                throw new ArgumentNullException(nameof(data));
            }

            await Task.Run(() =>
            {
                byte reportId = data.Length > 0 ? data[0] : (byte)0;
                IntPtr buffer = Marshal.AllocHGlobal(data.Length);

                try
                {
                    Marshal.Copy(data, 0, buffer, data.Length);
                    int res = IOHIDDeviceSetReport(_deviceHandle, 2, reportId, buffer, data.Length);
                    if (res != 0)
                    {
                        OnError?.Invoke(this, new HidErrorEventArgs(this, new Exception($"IOHIDDeviceSetReport (Feature) failed: {res}")));
                        return;
                    }
                }
                finally
                {
                    Marshal.FreeHGlobal(buffer);
                }
            });
        }

        public async Task<byte[]> GetFeatureReportAsync(byte reportId)
        {
            return await Task.Run(() =>
            {
                int length = FeatureReportByteLength;
                IntPtr buffer = Marshal.AllocHGlobal(length);

                try
                {
                    int res = IOHIDDeviceGetReport(_deviceHandle, 2, reportId, buffer, ref length);
                    if (res != 0)
                    {
                        OnError?.Invoke(this, new HidErrorEventArgs(this, new Exception($"IOHIDDeviceGetReport (Feature) failed: {res}")));
                        return Array.Empty<byte>();
                    }

                    byte[] outBuf = new byte[length];
                    Marshal.Copy(buffer, outBuf, 0, length);
                    return outBuf;
                }
                finally
                {
                    Marshal.FreeHGlobal(buffer);
                }
            });
        }

        public Task<byte[]> ReadInputReportAsync()
        {
            var tcs = new TaskCompletionSource<byte[]>();
            byte[] buffer = new byte[InputReportByteLength];

            InputReportCallback callback = null;

            callback = (ctx, result, sender, report, reportLength) =>
            {
                byte[] output = new byte[reportLength.ToInt32()];
                Array.Copy(report, output, output.Length);
                tcs.TrySetResult(output);
            };

            GCHandle.Alloc(callback);
            IOHIDDeviceRegisterInputReportCallback(_deviceHandle, buffer, (IntPtr)buffer.Length, callback, IntPtr.Zero);

            return tcs.Task;
        }


        public Task StartListeningAsync(CancellationToken ct)
        {
            if (_listeningTask != null && !_listeningTask.IsCompleted)
            {
                OnError?.Invoke(this, new HidErrorEventArgs(this, new NotSupportedException("Already listening on this device.")));
                return null;
            }

            _listeningCts = CancellationTokenSource.CreateLinkedTokenSource(ct);

            _listeningTask = Task.Run(() =>
            {
                try
                {
                    byte[] buffer = new byte[InputReportByteLength];
                    InputReportCallback callback = (ctx, result, sender, report, reportLength) =>
                    {
                        byte[] data = new byte[reportLength.ToInt32()];
                        Array.Copy(report, data, data.Length);

                        OnDataReceived?.Invoke(this, new HidDataReceivedEventArgs(this, data));
                    };

                    GCHandle.Alloc(callback);
                    IOHIDDeviceRegisterInputReportCallback(_deviceHandle, buffer, (IntPtr)buffer.Length, callback, IntPtr.Zero);
                    IOHIDDeviceScheduleWithRunLoop(_deviceHandle, CFRunLoopGetCurrent(), IntPtr.Zero);

                    while (!_listeningCts.Token.IsCancellationRequested)
                    {
                        CFRunLoopRun();
                    }

                    CFRunLoopStop(CFRunLoopGetCurrent());
                }
                catch (OperationCanceledException)
                {
                    return; // Exit gracefully if cancellation was requested
                }
                catch (NotSupportedException)
                {
                    OnError?.Invoke(this, new HidErrorEventArgs(this, new NotSupportedException("Reading input reports is not supported on this device.")));
                    _listeningCts.Cancel();
                }
                catch (Exception ex)
                {
                    // Handle exceptions during reading
                    if (_listeningCts.IsCancellationRequested)
                    {
                        // Exit if cancellation was requested
                    }
                    OnError?.Invoke(this, new HidErrorEventArgs(this, ex));
                }
            }, _listeningCts.Token);

            return _listeningTask;
        }

        private void InitializeDeviceInfo()
        {
            VendorId = GetIntProperty(kIOHIDVendorIDKey);
            ProductId = GetIntProperty(kIOHIDProductIDKey);
            Manufacturer = GetStringProperty(kIOHIDManufacturerKey);
            ProductName = GetStringProperty(kIOHIDProductKey);
            SerialNumber = GetStringProperty(kIOHIDSerialNumberKey);
        }

        private enum CFStringEncoding : uint
        {
            UTF8 = 0x08000100,
        }

        private ushort GetIntProperty(string key)
        {
            var cfKey = CFStringCreateWithCString(IntPtr.Zero, key, CFStringEncoding.UTF8);
            IntPtr val = IOHIDDeviceGetProperty(_deviceHandle, cfKey);
            CFRelease(cfKey);

            if (val == IntPtr.Zero)
            {
                return 0;
            }

            if (CFGetTypeID(val) == CFNumberGetTypeID())
            {
                int i = 0;
                bool success = CFNumberGetValue(val, CFNumberType.kCFNumberIntType, out i);
                if (success)
                {
                    return (ushort)i;
                }
            }

            return 0;
        }

        private string GetStringProperty(string key)
        {
            var cfKey = CFStringCreateWithCString(IntPtr.Zero, key, CFStringEncoding.UTF8);
            IntPtr val = IOHIDDeviceGetProperty(_deviceHandle, cfKey);
            CFRelease(cfKey);
            if (val == IntPtr.Zero)
            {
                return null;
            }

            string s = CFStringToString(val);
            CFRelease(val);
            return s;
        }

        private string CFStringToString(IntPtr cfStr)
        {
            if (cfStr == IntPtr.Zero)
            {
                return null;
            }

            IntPtr ptr = CFStringGetCStringPtr(cfStr, CFStringEncoding.UTF8);
            if (ptr != IntPtr.Zero)
            {
                // Manually read null-terminated UTF8 string from pointer
                return PtrToStringUTF8Manual(ptr);
            }

            int len = CFStringGetLength(cfStr);
            int max = CFStringGetMaximumSizeForEncoding(len, CFStringEncoding.UTF8) + 1;
            byte[] buf = new byte[max];

            bool success = CFStringGetCString(cfStr, buf, max, CFStringEncoding.UTF8);
            if (success)
            {
                return System.Text.Encoding.UTF8.GetString(buf).TrimEnd('\0');
            }

            return null;
        }

        private string PtrToStringUTF8Manual(IntPtr ptr)
        {
            if (ptr == IntPtr.Zero)
            {
                return null;
            }

            // Find length of null-terminated string
            int len = 0;
            while (Marshal.ReadByte(ptr, len) != 0)
            {
                len++;
            }

            if (len == 0)
            {
                return string.Empty;
            }

            byte[] buffer = new byte[len];
            Marshal.Copy(ptr, buffer, 0, len);
            return System.Text.Encoding.UTF8.GetString(buffer);
        }

        // Constants for IOHIDDevice properties
        private const string kIOHIDVendorIDKey = "VendorID";
        private const string kIOHIDProductIDKey = "ProductID";
        private const string kIOHIDManufacturerKey = "Manufacturer";
        private const string kIOHIDProductKey = "Product";
        private const string kIOHIDSerialNumberKey = "SerialNumber";

        [DllImport("/System/Library/Frameworks/IOKit.framework/IOKit")]
        static extern IntPtr IOHIDDeviceGetProperty(IntPtr dev, IntPtr key);
        [DllImport("/System/Library/Frameworks/IOKit.framework/IOKit")]
        static extern uint IOHIDDeviceOpen(IntPtr dev, uint opt);
        [DllImport("/System/Library/Frameworks/IOKit.framework/IOKit")]
        static extern uint IOHIDDeviceClose(IntPtr dev, uint opt);
        [DllImport("/System/Library/Frameworks/IOKit.framework/IOKit")]
        static extern int IOHIDDeviceSetReport(IntPtr dev, int type, byte id, IntPtr buf, int len);
        [DllImport("/System/Library/Frameworks/IOKit.framework/IOKit")]
        static extern int IOHIDDeviceGetReport(IntPtr dev, int type, byte id, IntPtr buf, ref int len);

        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        static extern IntPtr CFStringCreateWithCString(IntPtr alloc, string s, CFStringEncoding enc);
        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        static extern void CFRelease(IntPtr cf);
        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        static extern IntPtr CFRetain(IntPtr cf);
        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        static extern IntPtr CFStringGetCStringPtr(IntPtr str, CFStringEncoding enc);
        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        static extern int CFStringGetLength(IntPtr str);
        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        static extern int CFStringGetMaximumSizeForEncoding(int len, CFStringEncoding enc);
        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        static extern bool CFStringGetCString(IntPtr str, byte[] buf, int bufSize, CFStringEncoding enc);

        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        static extern IntPtr CFGetTypeID(IntPtr cf);
        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        static extern IntPtr CFNumberGetTypeID();
        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        static extern bool CFNumberGetValue(IntPtr num, CFNumberType type, out int value);

        [DllImport("/System/Library/Frameworks/IOKit.framework/IOKit")]
        private static extern void IOHIDDeviceRegisterInputReportCallback(
        IntPtr device,
        byte[] report,
        IntPtr reportLength,
        InputReportCallback callback,
        IntPtr context);

        [DllImport("/System/Library/Frameworks/IOKit.framework/IOKit")]
        private static extern void IOHIDDeviceScheduleWithRunLoop(
        IntPtr device,
        IntPtr runLoop,
        IntPtr runLoopMode);

        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        private static extern IntPtr CFRunLoopGetCurrent();

        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        private static extern void CFRunLoopRun();

        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        private static extern void CFRunLoopStop(IntPtr runLoop);

        // Input report callback delegate
        private delegate void InputReportCallback(
            IntPtr context,
            IntPtr result,
            IntPtr sender,
            byte[] report,
            IntPtr reportLength);

        // For managing lifetime of listening loop
        private CancellationTokenSource _listeningCts;
        private Task _listeningTask;


        private enum CFNumberType : int
        {
            kCFNumberIntType = 9
        }
    }
}
