﻿using EonaCat.HID.EventArguments;
using EonaCat.HID.Managers.Windows;
using Microsoft.Win32.SafeHandles;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using static EonaCat.HID.Managers.Windows.NativeMethods;

namespace EonaCat.HID
{
    // This file is part of the EonaCat project(s) which is released under the Apache License.
    // See the LICENSE file or go to https://EonaCat.com/license for full license details.

    internal sealed class HidWindows : IHid
    {
        private SafeFileHandle _deviceHandle;
        private FileStream _deviceStream;
        private IntPtr _preparsedData;

        private bool _isOpen;
        private readonly string _devicePath;

        public string DevicePath => _devicePath;
        public ushort VendorId { get; private set; }
        public ushort ProductId { get; private set; }
        public string SerialNumber { get; private set; }
        public string Manufacturer { get; private set; }
        public string ProductName { get; private set; }
        public int InputReportByteLength { get; private set; }
        public int OutputReportByteLength { get; private set; }
        public int FeatureReportByteLength { get; private set; }

        public IDictionary<string, object> Capabilities { get; } = new Dictionary<string, object>();

        public event EventHandler<HidDataReceivedEventArgs> OnDataReceived;
        public event EventHandler<HidErrorEventArgs> OnError;

        private CancellationTokenSource _listeningCts;

        private readonly object _lockObject = new object();

        public HidWindows(string devicePath)
        {
            _devicePath = devicePath ?? throw new ArgumentNullException(nameof(devicePath));
        }

        internal void Setup()
        {
            Open();
            LoadDeviceAttributes();
            Close();
        }

        private void LoadDeviceAttributes()
        {
            Manufacturer = GetStringDescriptor(HidD_GetManufacturerString);
            ProductName = GetStringDescriptor(HidD_GetProductString);
            SerialNumber = GetStringDescriptor(HidD_GetSerialNumberString);
            HidDeviceAttributes attr = GetDeviceAttributes();
            VendorId = attr.VendorID;
            ProductId = attr.ProductID;
        }

        /// <summary>
        /// Open the device for I/O
        /// </summary>
        public void Open()
        {
            if (_isOpen)
                return;

            FileAccess access = FileAccess.ReadWrite;
            SafeFileHandle handle = TryOpenDevice(GENERIC_READ | GENERIC_WRITE);

            if (handle == null || handle.IsInvalid)
            {
                handle = TryOpenDevice(GENERIC_READ);
                if (handle != null && !handle.IsInvalid)
                    access = FileAccess.Read;
            }

            if ((handle == null || handle.IsInvalid) && Environment.Is64BitOperatingSystem)
            {
                handle = TryOpenDevice(GENERIC_WRITE);
                if (handle != null && !handle.IsInvalid)
                    access = FileAccess.Write;
            }

            if (handle == null || handle.IsInvalid)
            {
                int err = Marshal.GetLastWin32Error();
                OnError?.Invoke(this, new HidErrorEventArgs(this, new Win32Exception(err, $"Cannot open device {_devicePath} with any access mode")));
                return;
            }

            _deviceHandle = handle;
            _deviceStream = new FileStream(_deviceHandle, access, 64, true);
            _isOpen = true;

            // HID descriptor parsing
            if (!HidD_GetPreparsedData(_deviceHandle.DangerousGetHandle(), out _preparsedData))
                throw new Win32Exception(Marshal.GetLastWin32Error(), "Failed HidD_GetPreparsedData");

            HIDP_CAPS caps;
            int res = HidP_GetCaps(_preparsedData, out caps);
            if (res == 0)
                throw new Win32Exception(res, "Failed HidP_GetCaps");

            InputReportByteLength = caps.InputReportByteLength;
            OutputReportByteLength = caps.OutputReportByteLength;
            FeatureReportByteLength = caps.FeatureReportByteLength;

            Capabilities["Usage"] = caps.Usage;
            Capabilities["UsagePage"] = caps.UsagePage;
            Capabilities["InputReportByteLength"] = InputReportByteLength;
            Capabilities["OutputReportByteLength"] = OutputReportByteLength;
            Capabilities["FeatureReportByteLength"] = FeatureReportByteLength;

            Manufacturer = GetStringDescriptor(HidD_GetManufacturerString);
            ProductName = GetStringDescriptor(HidD_GetProductString);
            SerialNumber = GetStringDescriptor(HidD_GetSerialNumberString);

            HidDeviceAttributes attr = GetDeviceAttributes();
            VendorId = attr.VendorID;
            ProductId = attr.ProductID;
        }


        private SafeFileHandle TryOpenDevice(int access)
        {
            var handle = CreateFile(_devicePath,
                access,
                FILE_SHARE_READ | FILE_SHARE_WRITE,
                IntPtr.Zero,
                OPEN_EXISTING,
                FILE_FLAG_OVERLAPPED,
                IntPtr.Zero);

            return handle;
        }

        private bool CanRead(SafeFileHandle handle)
        {
            try
            {
                using (var stream = new FileStream(handle, FileAccess.Read, 1, true))
                {
                    byte[] test = new byte[1];
                    stream.Read(test, 0, 1);
                    return true;
                }
            }
            catch { return false; }
        }

        private bool CanWrite(SafeFileHandle handle)
        {
            try
            {
                using (var stream = new FileStream(handle, FileAccess.Write, 1, true))
                {
                    byte[] test = new byte[1];
                    stream.Write(test, 0, 1);
                    return true;
                }
            }
            catch { return false; }
        }

        /// <summary>
        /// Close the device
        /// </summary>
        public void Close()
        {
            lock (_lockObject)
            {
                if (!_isOpen)
                {
                    return;
                }

                StopListening();

                if (_preparsedData != IntPtr.Zero)
                {
                    HidD_FreePreparsedData(_preparsedData);
                    _preparsedData = IntPtr.Zero;
                }

                _deviceStream?.Dispose();
                _deviceStream = null;

                _deviceHandle?.Dispose();
                _deviceHandle = null;

                _isOpen = false;
            }
        }
        private HidDeviceAttributes GetDeviceAttributes()
        {
            HidDeviceAttributes attr = new HidDeviceAttributes
            {
                Size = Marshal.SizeOf<HidDeviceAttributes>()
            };
            if (!HidD_GetAttributes(_deviceHandle.DangerousGetHandle(), ref attr))
            {
                throw new Win32Exception(Marshal.GetLastWin32Error(), "Failed HidD_GetAttributes");
            }
            return attr;
        }

        private string GetStringDescriptor(Func<IntPtr, IntPtr, int, bool> stringFunc)
        {
            var buffer = new byte[126 * 2]; // Unicode max buffer
            GCHandle gc = GCHandle.Alloc(buffer, GCHandleType.Pinned);
            try
            {
                bool success = stringFunc(_deviceHandle.DangerousGetHandle(), gc.AddrOfPinnedObject(), buffer.Length);
                if (!success)
                {
                    return null;
                }

                string str = Encoding.Unicode.GetString(buffer);
                int idx = str.IndexOf('\0');
                if (idx >= 0)
                {
                    str = str.Substring(0, idx);
                }

                return str;
            }
            finally
            {
                gc.Free();
            }
        }

        public void Dispose()
        {
            Close();
        }

        public async Task WriteOutputReportAsync(byte[] data)
        {
            if (!_isOpen)
            {
                OnError?.Invoke(this, new HidErrorEventArgs(this, new InvalidOperationException("Device not open")));
                return;
            }

            if (data == null || data.Length == 0)
            {
                OnError?.Invoke(this, new HidErrorEventArgs(this, new ArgumentNullException(nameof(data), "Data cannot be null or empty")));
                return;
            }

            try
            {
                await _deviceStream.WriteAsync(data, 0, data.Length);
                await _deviceStream.FlushAsync();
            }
            catch (Exception ex)
            {
                OnError?.Invoke(this, new HidErrorEventArgs(this, ex));
                throw;
            }
        }

        public async Task<byte[]> ReadInputReportAsync()
        {
            if (!_isOpen)
            {
                OnError?.Invoke(this, new HidErrorEventArgs(this, new InvalidOperationException("Device not open")));
                return Array.Empty<byte>();
            }

            return await Task.Run(async () =>
            {
                var buffer = new byte[InputReportByteLength];

                try
                {
                    int read = await _deviceStream.ReadAsync(buffer, 0, buffer.Length);
                    if (read == 0)
                    {
                        OnError?.Invoke(this, new HidErrorEventArgs(this, new IOException("No data read from device")));
                        return Array.Empty<byte>();
                    }

                    return buffer.Take(read).ToArray();
                }
                catch (Exception ex)
                {
                    OnError?.Invoke(this, new HidErrorEventArgs(this, ex));
                    throw;
                }
            });
        }

        public async Task SendFeatureReportAsync(byte[] data)
        {
            if (!_isOpen)
            {
                OnError?.Invoke(this, new HidErrorEventArgs(this, new InvalidOperationException("Device not open")));
                return;
            }

            if (data == null || data.Length == 0)
            {
                throw new ArgumentNullException(nameof(data));
            }

            await Task.Run(() =>
            {
                bool success = HidD_SetFeature(_deviceHandle.DangerousGetHandle(), data, data.Length);
                if (!success)
                {
                    var err = Marshal.GetLastWin32Error();
                    var ex = new Win32Exception(err, "HidD_SetFeature failed");
                    OnError?.Invoke(this, new HidErrorEventArgs(this, ex));
                    throw ex;
                }
            });
        }

        public async Task<byte[]> GetFeatureReportAsync(byte reportId)
        {
            if (!_isOpen)
            {
                OnError?.Invoke(this, new HidErrorEventArgs(this, new InvalidOperationException("Device not open")));
                return Array.Empty<byte>();
            }

            return await Task.Run(() =>
            {
                var buffer = new byte[FeatureReportByteLength];
                buffer[0] = reportId;

                bool success = HidD_GetFeature(_deviceHandle.DangerousGetHandle(), buffer, buffer.Length);
                if (!success)
                {
                    var err = Marshal.GetLastWin32Error();
                    var ex = new Win32Exception(err, "HidD_GetFeature failed");
                    OnError?.Invoke(this, new HidErrorEventArgs(this, ex));
                    return Array.Empty<byte>();
                }

                return buffer;
            });
        }

        /// <summary>
        /// Begin async reading loop raising OnDataReceived events on data input
        /// </summary>
        /// <param name="cancellationToken"></param>
        /// <returns></returns>
        public async Task StartListeningAsync(CancellationToken cancellationToken)
        {
            if (!_isOpen)
            {
                OnError?.Invoke(this, new HidErrorEventArgs(this, new NotSupportedException("Device is not open.")));
                return;
            }

            if (_listeningCts != null)
            {
                OnError?.Invoke(this, new HidErrorEventArgs(this, new NotSupportedException("Already listening on this device.")));
                return;
            }

            _listeningCts = CancellationTokenSource.CreateLinkedTokenSource(cancellationToken);

            try
            {
                while (!_listeningCts.Token.IsCancellationRequested)
                {
                    try
                    {
                        byte[] data = await ReadInputReportAsync(_listeningCts.Token);

                        if (data != null && data.Length > 0)
                        {
                            OnDataReceived?.Invoke(this, new HidDataReceivedEventArgs(this, data));
                        }
                    }
                    catch (OperationCanceledException)
                    {
                        break;
                    }
                    catch (NotSupportedException)
                    {
                        OnError?.Invoke(this, new HidErrorEventArgs(this, new NotSupportedException("Reading input reports is not supported on this device.")));
                        break; // Exit if reading is not supported
                    }
                    catch (Exception ex)
                    {
                        // Handle exceptions during reading
                        if (_listeningCts.IsCancellationRequested)
                        {
                            break; // Exit if cancellation was requested
                        }
                        OnError?.Invoke(this, new HidErrorEventArgs(this, ex));
                    }
                }
            }
            finally
            {
                _listeningCts.Dispose();
                _listeningCts = null;
            }
        }

        private Task<byte[]> ReadInputReportAsync(CancellationToken cancellationToken)
        {
            var tcs = new TaskCompletionSource<byte[]>(TaskCreationOptions.RunContinuationsAsynchronously);

            // Use overlapped IO pattern from FileStream
            var buffer = new byte[InputReportByteLength];
            _deviceStream.BeginRead(buffer, 0, buffer.Length, ar =>
            {
                try
                {
                    int bytesRead = _deviceStream.EndRead(ar);
                    if (bytesRead == 0)
                    {
                        tcs.SetResult(Array.Empty<byte>());
                    }
                    else
                    {
                        tcs.SetResult(buffer.Take(bytesRead).ToArray());
                    }
                }
                catch (Exception ex)
                {
                    tcs.SetException(ex);
                }
            }, null);

            cancellationToken.Register(() =>
            {
                tcs.TrySetCanceled();
            });

            return tcs.Task;
        }

        private void StopListening()
        {
            if (_listeningCts != null)
            {
                _listeningCts.Cancel();
                _listeningCts.Dispose();
                _listeningCts = null;
            }
        }
    }
}