﻿using EonaCat.HID.EventArguments;
using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using static EonaCat.HID.Managers.Mac.NativeMethods;

namespace EonaCat.HID.Managers.Mac
{
    // This file is part of the EonaCat project(s) which is released under the Apache License.
    // See the LICENSE file or go to https://EonaCat.com/license for full license details.

    internal sealed class HidManagerMac : IHidManager, IDisposable
    {
        private IntPtr _hidManager;
        private readonly IOHIDDeviceCallback _deviceAddedCallback;
        private readonly IOHIDDeviceCallback _deviceRemovedCallback;

        public event EventHandler<HidEventArgs> OnDeviceInserted;
        public event EventHandler<HidEventArgs> OnDeviceRemoved;

        public HidManagerMac()
        {
            _deviceAddedCallback = DeviceAddedCallback;
            _deviceRemovedCallback = DeviceRemovedCallback;

            _hidManager = IOHIDManagerCreate(IntPtr.Zero, 0);
            if (_hidManager == IntPtr.Zero)
            {
                throw new InvalidOperationException("Failed to create IOHIDManager");
            }

            CFRetain(_hidManager);

            IOHIDManagerScheduleWithRunLoop(
                _hidManager,
                CFRunLoopGetCurrent(),
                CFRunLoopModeDefault
            );

            if (IOHIDManagerOpen(_hidManager, 0) != 0)
            {
                throw new InvalidOperationException("Failed to open IOHIDManager");
            }

            IOHIDManagerRegisterDeviceMatchingCallback(_hidManager, _deviceAddedCallback, IntPtr.Zero);
            IOHIDManagerRegisterDeviceRemovalCallback(_hidManager, _deviceRemovedCallback, IntPtr.Zero);
        }

        public IEnumerable<IHid> Enumerate(ushort? vendorId = null, ushort? productId = null)
        {
            var devices = new List<IHid>();
            IntPtr cfSet = IOHIDManagerCopyDevices(_hidManager);
            if (cfSet == IntPtr.Zero)
            {
                return devices;
            }

            int count = CFSetGetCount(cfSet);
            IntPtr[] values = new IntPtr[count];
            CFSetGetValues(cfSet, values);
            foreach (var ptr in values)
            {
                var dev = new HidMac(ptr);
                dev.Setup();
                if (vendorId.HasValue && dev.VendorId != vendorId)
                {
                    continue;
                }

                if (productId.HasValue && dev.ProductId != productId)
                {
                    continue;
                }

                devices.Add(dev);
            }
            CFRelease(cfSet);
            return devices;
        }

        private void DeviceAddedCallback(IntPtr context, IntPtr result, IntPtr sender, IntPtr devicePtr)
        {
            if (devicePtr == IntPtr.Zero)
            {
                return; // Ignore null devices
            }

            // Create the device and invoke the event
            var device = new HidMac(devicePtr);
            device.Setup();
            OnDeviceInserted?.Invoke(this, new HidEventArgs(device));
        }

        private void DeviceRemovedCallback(IntPtr context, IntPtr result, IntPtr sender, IntPtr devicePtr)
        {
            if (devicePtr == IntPtr.Zero)
            {
                return; // Ignore null devices
            }

            OnDeviceRemoved?.Invoke(this, new HidEventArgs(null));
        }

        public void Dispose()
        {
            if (_hidManager != IntPtr.Zero)
            {
                IOHIDManagerUnscheduleFromRunLoop(_hidManager, CFRunLoopGetCurrent(), CFRunLoopModeDefault);
                IOHIDManagerClose(_hidManager, 0);
                CFRelease(_hidManager);
                _hidManager = IntPtr.Zero;
            }
        }
    }

    internal static class NativeMethods
    {
        public delegate void IOHIDDeviceCallback(IntPtr context, IntPtr result, IntPtr sender, IntPtr device);

        [DllImport("/System/Library/Frameworks/IOKit.framework/IOKit")]
        public static extern IntPtr IOHIDManagerCreate(IntPtr allocator, uint options);
        [DllImport("/System/Library/Frameworks/IOKit.framework/IOKit")]
        public static extern void IOHIDManagerScheduleWithRunLoop(IntPtr manager, IntPtr runLoop, IntPtr runLoopMode);
        [DllImport("/System/Library/Frameworks/IOKit.framework/IOKit")]
        public static extern void IOHIDManagerUnscheduleFromRunLoop(IntPtr manager, IntPtr runLoop, IntPtr runLoopMode);
        [DllImport("/System/Library/Frameworks/IOKit.framework/IOKit")]
        public static extern uint IOHIDManagerOpen(IntPtr manager, uint options);
        [DllImport("/System/Library/Frameworks/IOKit.framework/IOKit")]
        public static extern uint IOHIDManagerClose(IntPtr manager, uint options);
        [DllImport("/System/Library/Frameworks/IOKit.framework/IOKit")]
        public static extern IntPtr IOHIDManagerCopyDevices(IntPtr manager);
        [DllImport("/System/Library/Frameworks/IOKit.framework/IOKit")]
        public static extern void IOHIDManagerRegisterDeviceMatchingCallback(IntPtr manager, IOHIDDeviceCallback callback, IntPtr context);
        [DllImport("/System/Library/Frameworks/IOKit.framework/IOKit")]
        public static extern void IOHIDManagerRegisterDeviceRemovalCallback(IntPtr manager, IOHIDDeviceCallback callback, IntPtr context);

        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        public static extern IntPtr CFRunLoopGetCurrent();
        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        public static extern IntPtr CFRetain(IntPtr cf);
        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        public static extern void CFRelease(IntPtr cf);
        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        public static extern int CFSetGetCount(IntPtr cfset);
        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        public static extern void CFSetGetValues(IntPtr cfset, [Out] IntPtr[] values);
        [DllImport("/System/Library/Frameworks/CoreFoundation.framework/CoreFoundation")]
        public static extern IntPtr CFStringCreateWithCString(IntPtr alloc, string str, CFStringEncoding encoding);

        public static readonly IntPtr CFRunLoopModeDefault = CFStringCreateWithCString(
            IntPtr.Zero, "kCFRunLoopDefaultMode", CFStringEncoding.UTF8
        );

        public enum CFStringEncoding : uint
        {
            UTF8 = 0x08000100
        }
    }
}
